"""
.. module: historical.common.dynamodb
    :platform: Unix
    :copyright: (c) 2018 by Netflix Inc., see AUTHORS for more
    :license: Apache, see LICENSE for more details.
.. author:: Kevin Glisson <kglisson@netflix.com>
.. author:: Mike Grima <mgrima@netflix.com>
"""
import json
import logging

from deepdiff import DeepDiff
from boto3.dynamodb.types import TypeDeserializer

from historical.common.exceptions import DurableItemIsMissingException
from historical.constants import EVENT_TOO_BIG_FLAG

deser = TypeDeserializer()

log = logging.getLogger('historical')


'''
Important compatibility note:
-------------------------
- You can serialize and save a Durable object from a Current object if you run `remove_current_specific_fields`
- You can serialize, but NOT save a Current object from a Durable object if you run `remove_durable_specific_fields`
    ^^ This is by design. There should be no use case for saving a current item based on a durable configuration.
    
Thus, a Durable object is a subset of a Current object.
'''


def default_diff(latest_config, current_config):
    """Determine if two revisions have actually changed."""
    diff = DeepDiff(
        latest_config,
        current_config,
        ignore_order=True
    )
    return diff


def remove_global_dynamo_specific_fields(obj):
    """Remove all fields that are placed in by DynamoDB for global tables"""
    # https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/globaltables_HowItWorks.html
    obj.pop("aws:rep:deleting", None)
    obj.pop("aws:rep:updatetime", None)
    obj.pop("aws:rep:updateregion", None)

    return obj


def remove_current_specific_fields(obj):
    """Remove all fields that belong to the Current table -- that don't belong in the Durable table"""
    obj = remove_global_dynamo_specific_fields(obj)

    obj.pop("ttl", None)
    obj.pop("eventSource", None)

    return obj


def remove_durable_specific_fields(obj):
    """Remove all fields that belong to the Durable table -- that don't belong in the Durable table"""
    obj = remove_global_dynamo_specific_fields(obj)

    # Add here if any are required in the future.

    return obj


def modify_record(durable_model, current_revision, arn, event_time, diff_func):
    """Handles a DynamoDB MODIFY event type."""
    # We want the newest items first.
    # See: http://docs.aws.amazon.com/amazondynamodb/latest/developerguide/Query.html
    items = list(durable_model.query(
        arn,
        (durable_model.eventTime <= event_time),
        scan_index_forward=False,
        limit=1,
        consistent_read=True))

    if items:
        latest_revision = items[0]

        latest_config = latest_revision._get_json()[1]['attributes']
        current_config = current_revision._get_json()[1]['attributes']

        # Determine if there is truly a difference, disregarding Ephemeral Paths
        diff = diff_func(latest_config, current_config)
        if diff:
            log.debug(
                '[~] Difference found saving new revision to durable table. Arn: {} LatestConfig: {} '
                'CurrentConfig: {}'.format(arn, json.dumps(latest_config), json.dumps(current_config)))
            current_revision.save()
    else:
        current_revision.save()
        log.error('Got modify event but no current revision found. Arn: {arn}'.format(arn=arn))


def delete_differ_record(old_image, durable_model):
    """Handles a DynamoDB DELETE event type -- For the Differ."""
    data = {}
    for item in old_image:
        data[item] = deser.deserialize(old_image[item])

    data['configuration'] = {}
    # we give our own timestamps for TTL deletions
    del data['eventTime']
    durable_model(**data).save()
    log.debug('[+] Adding deletion marker.')


def get_full_current_object(record, current_model):
    """
    Utility method to fetch items from the Current table if they are too big for SNS/SQS.

    :param record:
    :param current_model:
    :return:
    """
    arn = record['dynamodb']['Keys']['arn']['S']
    log.debug('[-->] Item with ARN: {} was too big for SNS -- fetching it from the Current table...'.format(arn))
    item = list(current_model.query(arn))

    # If for whatever reason, the item *cannot* be found, then this record should be skipped over.
    # This will happen if this event came in and got processed right after the item was deleted
    # from the Current table. If so, then do nothing -- the deletion event will be processed later.
    if not item:
        return None

    # We need to place the real configuration data into the record so it can be deserialized into
    # the current model correctly:
    record['dynamodb']['NewImage']['configuration'] = item[0]._serialize()['attributes']['configuration']

    return record


def get_full_durable_object(record, durable_model):
    """
    Utility method to fetch items from the Durable table if they are too big for SNS/SQS.

    :param record:
    :param durable_model:
    :return:
    """
    arn = record['dynamodb']['Keys']['arn']['S']
    event_time = record['dynamodb']['NewImage']['eventTime']['S']
    log.debug('[-->] Item with ARN: {} was too big for SNS -- fetching it from the Durable table...'.format(arn))
    item = list(durable_model.query(arn, durable_model.eventTime == event_time))

    # It is not clear if this would ever be the case... We will consider this an error condition for now.
    if not item:
        log.error('[?] Item with ARN/Event Time: {}/{} was NOT found in the Durable table... This is odd.'.format(
            arn, event_time))
        raise DurableItemIsMissingException({"item_arn": arn, "event_time": event_time})

    # We need to place the real configuration data into the record so it can be deserialized into
    # the durable model correctly:
    record['dynamodb']['NewImage']['configuration'] = item[0]._serialize()['attributes']['configuration']

    return record


def deserialize_current_record_to_durable_model(record, current_model, durable_model):
    """
    Utility function that will take a dynamo event record and turn it into the proper pynamo object.

    This will properly deserialize the ugly Dynamo datatypes away.
    :param record:
    :param current_model:
    :param durable_model:
    :return:
    """
    # Was the item in question too big for SNS? If so, then we need to fetch the item from the current Dynamo table:
    if record.get(EVENT_TOO_BIG_FLAG):
        record = get_full_current_object(record, current_model)

        if not record:
            return None

    new_image = remove_current_specific_fields(record['dynamodb']['NewImage'])
    data = {}

    for item, value in new_image.items():
        # This could end up as loss of precision
        data[item] = deser.deserialize(value)

    return durable_model(**data)


def deserialize_current_record_to_current_model(record, current_model):
    """
    Utility function that will take a Dynamo event record and turn it into the proper Current Dynamo object.

    This will remove the "current table" specific fields, and properly deserialize the ugly Dynamo datatypes away.
    :param record:
    :param current_model:
    :return:
    """
    # Was the item in question too big for SNS? If so, then we need to fetch the item from the current Dynamo table:
    if record.get(EVENT_TOO_BIG_FLAG):
        record = get_full_current_object(record, current_model)

        if not record:
            return None

    new_image = remove_global_dynamo_specific_fields(record['dynamodb']['NewImage'])
    data = {}

    for item, value in new_image.items():
        # This could end up as loss of precision
        data[item] = deser.deserialize(value)

    return current_model(**data)


def deserialize_durable_record_to_durable_model(record, durable_model):
    """
    Utility function that will take a Dynamo event record and turn it into the proper Durable Dynamo object.

    This will properly deserialize the ugly Dynamo datatypes away.
    :param record:
    :param durable_model:
    :return:
    """
    # Was the item in question too big for SNS? If so, then we need to fetch the item from the current Dynamo table:
    if record.get(EVENT_TOO_BIG_FLAG):
        record = get_full_durable_object(record, durable_model)

        if not record:
            return None

    new_image = remove_global_dynamo_specific_fields(record['dynamodb']['NewImage'])
    data = {}

    for item, value in new_image.items():
        # This could end up as loss of precision
        data[item] = deser.deserialize(value)

    return durable_model(**data)


def deserialize_durable_record_to_current_model(record, current_model):
    """
    Utility function that will take a Durable Dynamo event record and turn it into the proper Current Dynamo object.

    This will properly deserialize the ugly Dynamo datatypes away.
    :param record:
    :param current_model:
    :return:
    """

    # Was the item in question too big for SNS? If so, then we need to fetch the item from the current Dynamo table:
    if record.get(EVENT_TOO_BIG_FLAG):
        # Try to get the data from the current table vs. grabbing the data from the Durable table:
        record = get_full_current_object(record, current_model)

        if not record:
            return None

    new_image = remove_durable_specific_fields(record['dynamodb']['NewImage'])
    data = {}

    for item, value in new_image.items():
        # This could end up as loss of precision
        data[item] = deser.deserialize(value)

    return current_model(**data)


def process_dynamodb_differ_record(record, current_model, durable_model, diff_func=None):
    """
    Processes a DynamoDB NewImage record (for Differ events).

    This will ONLY process the record if the record exists in one of the regions defined by the PROXY_REGIONS of
    the current Proxy function.
    """
    diff_func = diff_func or default_diff

    # Nothing special needs to be done for deletions as far as items that are too big for SNS are concerned.
    # This is because the deletion will remove the `configuration` field and save the item without it.
    if record['eventName'] == 'REMOVE':
        # We are *ONLY* tracking the deletions from the DynamoDB TTL service.
        # Why? Because when we process deletion records, we are first saving a new "empty" revision to the "Current"
        # table. The "empty" revision will then trigger this Lambda as a "MODIFY" event. Then, right after it saves
        # the "empty" revision, it will then delete the item from the "Current" table. At that point,
        # we have already saved the "deletion revision" to the "Durable" table. Thus, no need to process
        # the deletion events -- except for TTL expirations (which should never happen -- but if they do, you need
        # to investigate why...)
        if record.get('userIdentity'):
            if record['userIdentity']['type'] == 'Service':
                if record['userIdentity']['principalId'] == 'dynamodb.amazonaws.com':
                    log.error('We received a TTL delete. Old Image: {}'.format(record['dynamodb']['OldImage']))
                    old_image = remove_current_specific_fields(record['dynamodb']['OldImage'])
                    delete_differ_record(old_image, durable_model)

    if record['eventName'] in ['INSERT', 'MODIFY']:
        arn = record['dynamodb']['Keys']['arn']['S']
        current_revision = deserialize_current_record_to_durable_model(record, current_model, durable_model)

        if not current_revision:
            log.error("[?] Received item too big for SNS, and was not able to find the original item with ARN: {}"
                      .format(arn))
            return

        if record['eventName'] == 'INSERT':
            current_revision.save()
            log.debug('[+] Saving new revision to durable table.')

        elif record['eventName'] == 'MODIFY':
            modify_record(durable_model, current_revision, arn, current_revision.eventTime, diff_func)



"""
.. module: historical.models
    :platform: Unix
    :copyright: (c) 2017 by Netflix Inc., see AUTHORS for more
    :license: Apache, see LICENSE for more details.
.. author:: Kevin Glisson <kglisson@netflix.com>
.. author:: Mike Grima <mgrima@netflix.com>
"""
import time
from datetime import datetime

from marshmallow import Schema, fields
from pynamodb.attributes import UnicodeAttribute, MapAttribute, NumberAttribute

from historical.attributes import EventTimeAttribute


TTL_EXPIRY = 86400  # 24 Hours in seconds

EPHEMERAL_PATHS = [
]


def default_ttl():
    return int(time.time() + TTL_EXPIRY)


def default_event_time():
    return datetime.utcnow().replace(tzinfo=None, microsecond=0).isoformat() + "Z"


class DurableHistoricalModel(object):
    eventTime = EventTimeAttribute(range_key=True, default=default_event_time)


class CurrentHistoricalModel(object):
    eventTime = EventTimeAttribute(default=default_event_time)
    ttl = NumberAttribute(default=default_ttl())
    eventSource = UnicodeAttribute()


class AWSHistoricalMixin(object):
    arn = UnicodeAttribute(hash_key=True)
    accountId = UnicodeAttribute()
    userIdentity = MapAttribute(null=True)
    principalId = UnicodeAttribute(null=True)
    configuration = MapAttribute(null=True)
    userAgent = UnicodeAttribute(null=True)
    sourceIpAddress = UnicodeAttribute(null=True)
    requestParameters = MapAttribute(null=True)


class HistoricalPollingEventDetail(Schema):
    # You must replace these:
    event_source = fields.Str(dump_to="eventSource", load_from="eventSource", required=True)
    event_name = fields.Str(dump_to="eventName", load_from="eventName", required=True)
    request_parameters = fields.Dict(dump_to="requestParameters", load_from="requestParameters", required=True)

    event_time = fields.Str(dump_to="eventTime", load_from="eventTime", required=True,
                            default=default_event_time, missing=default_event_time)


class HistoricalPollingBaseModel(Schema):
    version = fields.Str(required=True)
    account = fields.Str(required=True)

    detail_type = fields.Str(load_from="detail-type", dump_to="detail-type", required=True,
                             missing='Historical Polling Event', default='Historical Polling Event')
    source = fields.Str(required=True, missing='historical', default='historical')
    time = fields.Str(required=True, default=default_event_time, missing=default_event_time)

    # You must replace this:
    detail = fields.Nested(HistoricalPollingEventDetail, required=True)


class HistoricalPollerTaskEventModel(Schema):
    account_id = fields.Str(required=True)
    region = fields.Str(required=True)

    def serialize_me(self, account_id, region):
        return self.dumps({
            "account_id": account_id,
            "region": region
        }).data



"""
.. module: historical.s3.models
    :platform: Unix
    :copyright: (c) 2017 by Netflix Inc., see AUTHORS for more
    :license: Apache, see LICENSE for more details.
.. author:: Mike Grima <mgrima@netflix.com>
"""
from marshmallow import Schema, fields, post_dump
from pynamodb.models import Model
from pynamodb.attributes import UnicodeAttribute, NumberAttribute, MapAttribute
from pynamodb.indexes import GlobalSecondaryIndex, AllProjection

from historical.constants import CURRENT_REGION
from historical.models import DurableHistoricalModel, CurrentHistoricalModel, AWSHistoricalMixin, \
    HistoricalPollingEventDetail, HistoricalPollingBaseModel


class S3Model(object):
    BucketName = UnicodeAttribute()
    Region = UnicodeAttribute()
    Tags = MapAttribute()


class DurableS3Model(Model, DurableHistoricalModel, AWSHistoricalMixin, S3Model):
    class Meta:
        table_name = 'HistoricalS3DurableTable'
        region = CURRENT_REGION


class CurrentS3Model(Model, CurrentHistoricalModel, AWSHistoricalMixin, S3Model):
    class Meta:
        table_name = 'HistoricalS3CurrentTable'
        region = CURRENT_REGION


class ViewIndex(GlobalSecondaryIndex):
    class Meta:
        projection = AllProjection()
        region = CURRENT_REGION

    view = NumberAttribute(default=0, hash_key=True)


class S3PollingRequestParamsModel(Schema):
    bucket_name = fields.Str(dump_to="bucketName", load_from="bucketName", required=True)
    creation_date = fields.Str(dump_to="creationDate", load_from="creationDate", required=True)


class S3PollingEventDetail(HistoricalPollingEventDetail):
    request_parameters = fields.Nested(S3PollingRequestParamsModel, dump_to="requestParameters",
                                       load_from="requestParameters", required=True)

    event_source = fields.Str(load_only=True, load_from="eventSource", required=True)
    event_name = fields.Str(load_only=True, load_from="eventName", required=True)

    @post_dump
    def add_required_s3_polling_data(self, data):
        data["eventSource"] = "historical.s3.poller"
        data["eventName"] = "DescribeBucket"

        return data


class S3PollingEventModel(HistoricalPollingBaseModel):
    detail = fields.Nested(S3PollingEventDetail, required=True)

    version = fields.Str(load_only=True, required=True)

    @post_dump()
    def dump_s3_polling_event_data(self, data):
        data["version"] = "1"

        return data

    def serialize_me(self, account, bucket_details):
        return self.dumps({
            "account": account,
            "detail": {
                "request_parameters": {
                    "bucket_name": bucket_details["Name"],
                    "creation_date": bucket_details["CreationDate"].replace(
                        tzinfo=None, microsecond=0).isoformat() + "Z"
                }
            }
        }).data


s3_polling_schema = S3PollingEventModel(strict=True)



"""
.. module: historical.common.proxy
    :platform: Unix
    :copyright: (c) 2018 by Netflix Inc., see AUTHORS for more
    :license: Apache, see LICENSE for more details.
.. author:: Mike Grima <mgrima@netflix.com>
"""
import logging
import json
import math
import os
import sys

import boto3
from retrying import retry

from raven_python_lambda import RavenLambdaWrapper

from historical.common.exceptions import MissingProxyConfigurationException
from historical.common.sqs import produce_events
from historical.constants import CURRENT_REGION, REGION_ATTR, PROXY_REGIONS, EVENT_TOO_BIG_FLAG

log = logging.getLogger('historical')


@retry(stop_max_attempt_number=4, wait_exponential_multiplier=1000, wait_exponential_max=1000)
def _publish_sns_message(client, blob, topic_arn):
    client.publish(TopicArn=topic_arn, Message=blob)


def shrink_blob(record):
    """
    Makes a shrunken blob to be sent to SNS/SQS (due to the 256KB size limitations of SNS/SQS messages).
    This will essentially remove the "configuration" field such that the size of the SNS/SQS message remains under 256KB.
    :param record:
    :return:
    """
    item = {
        "eventName": record["eventName"],
        EVENT_TOO_BIG_FLAG: True
    }

    # To handle TTLs (if they happen)
    if record.get("userIdentity"):
        item["userIdentity"] = record["userIdentity"]

    # Remove the 'configuration' field from new and old images if applicable:
    if record['dynamodb'].get('NewImage'):
        record['dynamodb']['NewImage'].pop('configuration', None)

    if record['dynamodb'].get('OldImage'):
        record['dynamodb']['OldImage'].pop('configuration', None)

    item['dynamodb'] = record['dynamodb']

    return item


@RavenLambdaWrapper()
def handler(event, context):
    """Historical S3 DynamoDB Stream Forwarder (the 'Proxy').

    Passes events from the Historical DynamoDB stream and passes it to SNS or SQS for additional events to trigger.

    You can optionally use SNS or SQS. It is preferable to use SNS -> SQS, but in some cases, such as the Current stream
    to the Differ, this will make use of SQS to directly feed into the differ for performance purposes.
    """
    queue_url = os.environ.get('PROXY_QUEUE_URL')
    topic_arn = os.environ.get('PROXY_TOPIC_ARN')

    if not queue_url and not topic_arn:
        raise MissingProxyConfigurationException('[X] Must set the `PROXY_QUEUE_URL` or the `PROXY_TOPIC_ARN` vars.')

    items_to_ship = [make_proper_record(record) for record in event['Records']]

    # SQS:
    if queue_url:
        produce_events(items_to_ship, queue_url)

    # SNS:
    else:
        client = boto3.client("sns", region_name=CURRENT_REGION)
        for i in items_to_ship:
            _publish_sns_message(client, i, topic_arn)


def make_proper_record(record):
    """Prepares and ships an individual record over to SNS/SQS for future processing.

    :param record:
    :return:
    """
    # We should NOT be processing this if the item in question does not
    # reside in the PROXY_REGIONS
    for img in ['NewImage', 'OldImage']:
        if record['dynamodb'].get(img):
            if record['dynamodb'][img][REGION_ATTR]['S'] not in PROXY_REGIONS:
                log.debug("[/] Not processing record -- record event took place in: {}".format(
                    record['dynamodb'][img][REGION_ATTR]['S']))
                return

    blob = json.dumps(record)

    size = math.ceil(sys.getsizeof(blob) / 1024)
    if size >= 256:
        # Need to send over a smaller blob to inform the recipient that it needs to go out and
        # fetch the item from the Historical table!
        blob = json.dumps(shrink_blob(record))

    return blob



"""
.. module: historical.vpc.poller
    :platform: Unix
    :copyright: (c) 2017 by Netflix Inc., see AUTHORS for more
    :license: Apache, see LICENSE for more details.
.. author:: Kevin Glisson <kglisson@netflix.com>
.. author:: Mike Grima <mgrima@netflix.com>
"""
import os
import logging

from botocore.exceptions import ClientError

from raven_python_lambda import RavenLambdaWrapper
from cloudaux.aws.ec2 import describe_vpcs

from historical.constants import POLL_REGIONS, HISTORICAL_ROLE, LOGGING_LEVEL
from historical.common.util import deserialize_records
from historical.vpc.models import vpc_polling_schema
from historical.models import HistoricalPollerTaskEventModel
from historical.common.accounts import get_historical_accounts
from historical.common.sqs import produce_events, get_queue_url

logging.basicConfig()
log = logging.getLogger("historical")
log.setLevel(LOGGING_LEVEL)


@RavenLambdaWrapper()
def poller_tasker_handler(event, context):
    """
    Historical VPC Poller Tasker.

    The Poller is run at a set interval in order to ensure that changes do not go undetected by Historical.

    Historical pollers generate `polling events` which simulate changes. These polling events contain configuration
    data such as the account/region defining where the collector should attempt to gather data from.

    This is the entry point. This will task subsequent Poller lambdas to list all of a given resource in a select few
    AWS accounts.
    """
    log.debug('[@] Running Poller Tasker...')

    queue_url = get_queue_url(os.environ.get('POLLER_TASKER_QUEUE_NAME', 'HistoricalVPCPollerTasker'))
    poller_task_schema = HistoricalPollerTaskEventModel()

    events = []
    for account in get_historical_accounts():
        for region in POLL_REGIONS:
            events.append(poller_task_schema.serialize_me(account['id'], region))

    try:
        produce_events(events, queue_url)
    except ClientError as e:
        log.error('[X] Unable to generate poller tasker events! Reason: {reason}'.format(reason=e))

    log.debug('[@] Finished tasking the pollers.')


@RavenLambdaWrapper()
def poller_processor_handler(event, context):
    """
    Historical Security Group Poller Processor.

    This will receive events from the Poller Tasker, and will list all objects of a given technology for an
    account/region pair. This will generate `polling events` which simulate changes. These polling events contain
    configuration data such as the account/region defining where the collector should attempt to gather data from.
    """
    log.debug('[@] Running Poller...')

    queue_url = get_queue_url(os.environ.get('POLLER_QUEUE_NAME', 'HistoricalVPCPoller'))

    records = deserialize_records(event['Records'])

    for record in records:
        # Skip accounts that have role assumption errors:
        try:
            vpcs = describe_vpcs(
                account_number=record['account_id'],
                assume_role=HISTORICAL_ROLE,
                region=record['region']
            )

            events = [vpc_polling_schema.serialize(record['account_id'], v) for v in vpcs]
            produce_events(events, queue_url)
            log.debug('[@] Finished generating polling events. Account: {}/{} '
                      'Events Created: {}'.format(record['account_id'], record['region'], len(events)))
        except ClientError as e:
            log.error('[X] Unable to generate events for account/region. Account Id/Region: {account_id}/{region}'
                      ' Reason: {reason}'.format(account_id=record['account_id'], region=record['region'], reason=e))

        log.debug('[@] Finished generating polling events. Events Created: {}'.format(len(record['account_id'])))



import os
import logging
import click
import click_log
from cookiecutter.main import cookiecutter

from historical.__about__ import __version__

log = logging.getLogger('historical')
click_log.basic_config(log)


@click.group()
@click_log.simple_verbosity_option(log)
@click.version_option(version=__version__)
def cli():
    """Historical commandline for managing historical functions."""
    pass


@cli.command()
def new():
    """Creates a new historical technology."""
    dir_path = os.path.dirname(os.path.realpath(__file__))
    cookiecutter(os.path.join(dir_path, 'historical-cookiecutter/'))



"""
.. module: historical.commom.util
    :platform: Unix
    :copyright: (c) 2018 by Netflix Inc., see AUTHORS for more
    :license: Apache, see LICENSE for more details.
.. author:: Mike Grima <mgrima@netflix.com>
"""
import json


def deserialize_records(records):
    """
    This properly deserializes records depending on where they came from:
        - SQS
        - SNS
    """
    native_records = []
    for r in records:
        parsed = json.loads(r['body'])

        # Is this a DynamoDB stream event?
        if isinstance(parsed, str):
            native_records.append(json.loads(parsed))

        # Is this from SNS (cross-region request -- SNS messages wrapped in SQS message) -- or an SNS proxied message?
        elif parsed.get('EventSource') == 'aws:sns' and parsed.get('Message'):
            native_records.append(json.loads(parsed['Message']))

        else:
            native_records.append(parsed)

    return native_records



"""
.. module: historical.common.extensions
    :platform: Unix
    :copyright: (c) 2018 by Netflix Inc., see AUTHORS for more
    :license: Apache, see LICENSE for more details.
.. author:: Kevin Glisson <kglisson@netflix.com>
"""
from jinja2.ext import Extension


def titlecase(f):
    """Transforms a string to titlecase."""
    return "".join([x.title() for x in f.split('_')])


class HistoricalExtension(Extension):
    def __init__(self, environment):
        super(HistoricalExtension, self).__init__(environment)
        environment.filters['titlecase'] = titlecase



"""
.. module: historical.common.exceptions
    :platform: Unix
    :copyright: (c) 2018 by Netflix Inc., see AUTHORS for more
    :license: Apache, see LICENSE for more details.
.. author:: Mike Grima <mgrima@netflix.com>
"""


class DurableItemIsMissingException(Exception):
    pass


class MissingProxyConfigurationException(Exception):
    pass



"""
.. module: historical.security_group.differ
    :platform: Unix
    :copyright: (c) 2017 by Netflix Inc., see AUTHORS for more
    :license: Apache, see LICENSE for more details.
.. author:: Kevin Glisson <kglisson@netflix.com>
"""
import logging

from raven_python_lambda import RavenLambdaWrapper

from historical.common.dynamodb import process_dynamodb_differ_record
from historical.common.util import deserialize_records
from historical.security_group.models import DurableSecurityGroupModel, CurrentSecurityGroupModel
from historical.constants import LOGGING_LEVEL

logging.basicConfig()
log = logging.getLogger('historical')
log.setLevel(LOGGING_LEVEL)


@RavenLambdaWrapper()
def handler(event, context):
    """
    Historical security group event differ.

    Listens to the Historical current table and determines if there are differences that need to be persisted in the
    historical record.
    """
    # De-serialize the records:
    records = deserialize_records(event['Records'])

    for record in records:
        process_dynamodb_differ_record(record, CurrentSecurityGroupModel, DurableSecurityGroupModel)
